/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.ui.rest.model.artifacts.browse.treebrowser.nodes.actions;

import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.descriptor.config.CentralConfigDescriptor;
import org.artifactory.descriptor.download.FolderDownloadConfigDescriptor;
import org.artifactory.descriptor.repo.RemoteRepoDescriptor;
import org.artifactory.repo.RepoPath;
import org.artifactory.repo.RepoPathFactory;
import org.mockito.Mockito;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertNotNull;
import static org.testng.Assert.assertNull;

/**
 * @author Rotem Kfir
 */
@Test
public class TreeNodeActionsPopulatorTest {

    private TreeNodeActionsPopulator populator;
    private RepositoryService repoService;
    private AuthorizationService authService;
    private CentralConfigService centralConfig;

    @BeforeMethod
    public void setup() {
        repoService = Mockito.mock(RepositoryService.class);
        authService = Mockito.mock(AuthorizationService.class);
        centralConfig = Mockito.mock(CentralConfigService.class);
        populator = new TreeNodeActionsPopulator(repoService, authService, centralConfig, null, null);
    }

    @Test
    public void testSuccessfulDownloadFolderForAnonymous() throws Exception {
        enableFolderDownload(true);
        makeTheUserAnonymousWithReadPermissions();
        TabOrAction action = populator.downloadFolder(null);
        assertNotNull(action);
    }

    @Test
    public void testDownloadFolderNotEnabledForAnonymous() throws Exception {
        enableFolderDownload(false);
        makeTheUserAnonymousWithReadPermissions();
        TabOrAction action = populator.downloadFolder(null);
        assertNull(action);
    }

    @Test(dataProvider = "getOptions")
    public void testUserCanDownloadFromRemote(boolean userCanDeploy, boolean fileExistInCache,
            boolean remoteCacheEnabled, boolean downloadShouldBeAvailable) {
        setupRepoService(userCanDeploy, fileExistInCache, remoteCacheEnabled);

        TabsAndActions tabsAndActions = new TabsAndActions();
        tabsAndActions.setRepoType("remote");
        tabsAndActions.setRepoPath("test-remote/item");

        populator.populateForFile(tabsAndActions);

        if (downloadShouldBeAvailable) {
            assertEquals(tabsAndActions.getActions().size(), 1);
            assertEquals(tabsAndActions.getActions().get(0).getName(), "Download");
        } else {
            assertEquals(tabsAndActions.getActions().size(), 0);
        }
    }

    private void enableFolderDownload(boolean forAnonymous) {
        FolderDownloadConfigDescriptor folderDescriptor = new FolderDownloadConfigDescriptor();
        folderDescriptor.setEnabled(true);
        folderDescriptor.setEnabledForAnonymous(forAnonymous);
        CentralConfigDescriptor descriptor = Mockito.mock(CentralConfigDescriptor.class);
        Mockito.when(descriptor.getFolderDownloadConfig()).thenReturn(folderDescriptor);
        Mockito.when(centralConfig.getDescriptor()).thenReturn(descriptor);
    }

    private void makeTheUserAnonymousWithReadPermissions() {
        Mockito.when(authService.canRead(Mockito.any())).thenReturn(true);
        Mockito.when(authService.isAnonymous()).thenReturn(true);
    }

    private void setupRepoService(boolean userCanDeploy, boolean fileExistInCache, boolean remoteCacheEnabled) {
        RepoPath remotePath = RepoPathFactory.create("test-remote/item");
        RepoPath remoteCachePath = RepoPathFactory.create("test-remote-cache");
        RemoteRepoDescriptor remote = new RemoteRepoDescriptor() {
        };
        remote.setKey("test-remote");
        remote.setStoreArtifactsLocally(remoteCacheEnabled);
        Mockito.when(repoService.remoteRepoDescriptorByKey("test-remote")).thenReturn(remote);
        Mockito.when(repoService.exists(remotePath)).thenReturn(fileExistInCache);
        Mockito.when(authService.canDeploy(RepoPathFactory.create("test-remote-cache/item"))).thenReturn(userCanDeploy);
    }

    @DataProvider
    public static Object[][] getOptions() {
        return new Object[][] {
                //userCanDeploy, fileExistInCache, remoteCacheEnabled, downloadShouldBeAvailable
                {false, false, false, true},
                {false, false, true, false},
                {false, true, false, true},
                {false, true, true, true},
                {true, false, false, true},
                {true, false, true, true},
                {true, true, true, true}
        };
    }
}